@@ -1,7 +1,7 @@
 /*
  * bibprogs.h
  *
- * Copyright (c) Chris Putnam 2008-2017
+ * Copyright (c) Chris Putnam 2008-2020
  *
  * Program and source code released under the GPL version 2
  *
@@ -11,6 +11,6 @@
 
 #include "bibutils.h"
 
-extern void bibprog( int argc, char *argv[], param *p );
+void bibprog( int argc, char *argv[], param *p );
 
 #endif
