@@ -1,8 +1,8 @@
 /*
  * biblatexin.c
  *
- * Copyright (c) Chris Putnam 2008-2017
- * Copyright (c) Johannes Wilm 2010-2017
+ * Copyright (c) Chris Putnam 2008-2020
+ * Copyright (c) Johannes Wilm 2010-2020
  *
  * Program and source code released under the GPL version 2
  *
@@ -17,6 +17,7 @@
 #include "utf8.h"
 #include "str_conv.h"
 #include "fields.h"
+#include "latex_parse.h"
 #include "slist.h"
 #include "name.h"
 #include "reftypes.h"
@@ -34,38 +35,43 @@ static slist replace = { 0, 0, 0, NULL };
 *****************************************************/
 
 static int  biblatexin_convertf( fields *bibin, fields *info, int reftype, param *p );
-static int  biblatexin_processf( fields *bibin, char *data, char *filename, long nref, param *p );
+static int  biblatexin_processf( fields *bibin, const char *data, const char *filename, long nref, param *p );
 static int  biblatexin_cleanf( bibl *bin, param *p );
 static int  biblatexin_readf( FILE *fp, char *buf, int bufsize, int *bufpos, str *line, str *reference, int *fcharset );
-static int  biblatexin_typef( fields *bibin, char *filename, int nrefs, param *p );
+static int  biblatexin_typef( fields *bibin, const char *filename, int nrefs, param *p );
 
-void
-biblatexin_initparams( param *p, const char *progname )
+int
+biblatexin_initparams( param *pm, const char *progname )
 {
-	p->readformat       = BIBL_BIBLATEXIN;
-	p->charsetin        = BIBL_CHARSET_DEFAULT;
-	p->charsetin_src    = BIBL_SRC_DEFAULT;
-	p->latexin          = 1;
-	p->xmlin            = 0;
-	p->utf8in           = 0;
-	p->nosplittitle     = 0;
-	p->verbose          = 0;
-	p->addcount         = 0;
-	p->output_raw       = 0;
+	pm->readformat       = BIBL_BIBLATEXIN;
+	pm->charsetin        = BIBL_CHARSET_DEFAULT;
+	pm->charsetin_src    = BIBL_SRC_DEFAULT;
+	pm->latexin          = 1;
+	pm->xmlin            = 0;
+	pm->utf8in           = 0;
+	pm->nosplittitle     = 0;
+	pm->verbose          = 0;
+	pm->addcount         = 0;
+	pm->output_raw       = 0;
 
-	p->readf    = biblatexin_readf;
-	p->processf = biblatexin_processf;
-	p->cleanf   = biblatexin_cleanf;
-	p->typef    = biblatexin_typef;
-	p->convertf = biblatexin_convertf;
-	p->all      = biblatex_all;
-	p->nall     = biblatex_nall;
+	pm->readf    = biblatexin_readf;
+	pm->processf = biblatexin_processf;
+	pm->cleanf   = biblatexin_cleanf;
+	pm->typef    = biblatexin_typef;
+	pm->convertf = biblatexin_convertf;
+	pm->all      = biblatex_all;
+	pm->nall     = biblatex_nall;
 
-	slist_init( &(p->asis) );
-	slist_init( &(p->corps) );
+	slist_init( &(pm->asis) );
+	slist_init( &(pm->corps) );
 
-	if ( !progname ) p->progname = NULL;
-	else p->progname = strdup( progname );
+	if ( !progname ) pm->progname = NULL;
+	else {
+		pm->progname = strdup( progname );
+		if ( !pm->progname ) return BIBL_ERR_MEMERR;
+	}
+
+	return BIBL_OK;
 }
 
 /*****************************************************
@@ -97,7 +103,7 @@ static int
 biblatexin_readf( FILE *fp, char *buf, int bufsize, int *bufpos, str *line, str *reference, int *fcharset )
 {
 	int haveref = 0;
-	char *p;
+	const char *p;
 	while ( haveref!=2 && readmore( fp, buf, bufsize, bufpos, line ) ) {
 		if ( line->len == 0 ) continue; /* blank line */
 		p = &(line->data[0]);
@@ -112,7 +118,6 @@ biblatexin_readf( FILE *fp, char *buf, int bufsize, int *bufpos, str *line, str
 			str_addchar( reference, '\n' );
 			str_empty( line );
 		} else if ( !haveref ) str_empty( line );
-	
 	}
 	*fcharset = CHARSET_UNKNOWN;
 	return haveref;
@@ -122,8 +127,8 @@ biblatexin_readf( FILE *fp, char *buf, int bufsize, int *bufpos, str *line, str
  PUBLIC: int biblatexin_processf()
 *****************************************************/
 
-static char *
-process_biblatextype( char *p, str *type )
+static const char *
+process_biblatextype( const char *p, str *type )
 {
 	str tmp;
 	str_init( &tmp );
@@ -141,10 +146,10 @@ process_biblatextype( char *p, str *type )
 	return p;
 }
 
-static char *
-process_biblatexid( char *p, str *id )
+static const char *
+process_biblatexid( const char *p, str *id )
 {
-	char *start_p = p;
+	const char *start_p = p;
 	str tmp;
 
 	str_init( &tmp );
@@ -170,19 +175,20 @@ process_biblatexid( char *p, str *id )
 	return skip_ws( p );
 }
 
-static char *
-biblatex_tag( char *p, str *tag )
+static const char *
+biblatex_tag( const char *p, str *tag )
 {
 	p = str_cpytodelim( tag, skip_ws( p ), "= \t\r\n", 0 );
 	return skip_ws( p );
 }
 
-static char *
-biblatex_data( char *p, fields *bibin, slist *tokens, long nref, param *pm )
+static const char *
+biblatex_data( const char *p, fields *bibin, slist *tokens, long nref, param *pm )
 {
 	unsigned int nbracket = 0, nquotes = 0;
-	char *startp = p;
-	str tok, *s;
+	const char *startp = p;
+	int status;
+	str tok;
 
 	str_init( &tok );
 	while ( p && *p ) {
@@ -194,18 +200,18 @@ biblatex_data( char *p, fields *bibin, slist *tokens, long nref, param *pm )
 			nquotes = !nquotes;
 			str_addchar( &tok, *p );
 			if ( !nquotes ) {
-				s = slist_add( tokens, &tok );
-				if ( !s ) { p = NULL; goto outerr; }
+				status = slist_add( tokens, &tok );
+				if ( status!=SLIST_OK ) { p = NULL; goto outerr; }
 				str_empty( &tok );
 			}
 		} else if ( *p=='#' && !nquotes && !nbracket ) {
 			if ( str_has_value( &tok ) ) {
-				s = slist_add( tokens, &tok );
-				if ( !s ) { p = NULL; goto outerr; }
+				status = slist_add( tokens, &tok );
+				if ( status!=SLIST_OK ) { p = NULL; goto outerr; }
 			}
 			str_strcpyc( &tok, "#" );
-			s = slist_add( tokens, &tok );
-			if ( !s ) { p = NULL; goto outerr; }
+			status = slist_add( tokens, &tok );
+			if ( status!=SLIST_OK ) { p = NULL; goto outerr; }
 			str_empty( &tok );
 		} else if ( *p=='{' && !nquotes && ( p==startp || *(p-1)!='\\' ) ) {
 			nbracket++;
@@ -214,8 +220,8 @@ biblatex_data( char *p, fields *bibin, slist *tokens, long nref, param *pm )
 			nbracket--;
 			str_addchar( &tok, *p );
 			if ( nbracket==0 ) {
-				s = slist_add( tokens, &tok );
-				if ( !s ) { p = NULL; goto outerr; }
+				status = slist_add( tokens, &tok );
+				if ( status!=SLIST_OK ) { p = NULL; goto outerr; }
 				str_empty( &tok );
 			}
 		} else if ( !is_ws( *p ) || nquotes || nbracket ) {
@@ -230,8 +236,8 @@ biblatex_data( char *p, fields *bibin, slist *tokens, long nref, param *pm )
 			}
 		} else if ( is_ws( *p ) ) {
 			if ( str_has_value( &tok ) ) {
-				s = slist_add( tokens, &tok );
-				if ( !s ) { p = NULL; goto outerr; }
+				status = slist_add( tokens, &tok );
+				if ( status!=SLIST_OK ) { p = NULL; goto outerr; }
 				str_empty( &tok );
 			}
 		}
@@ -245,8 +251,8 @@ out:
 		fprintf( stderr, "%s: Mismatch in number of quotes in reference %ld\n", pm->progname, nref );
 	}
 	if ( str_has_value( &tok ) ) {
-		s = slist_add( tokens, &tok );
-		if ( !s ) p = NULL;
+		status = slist_add( tokens, &tok );
+		if ( status!=SLIST_OK ) p = NULL;
 	}
 outerr:
 	str_free( &tok );
@@ -310,7 +316,7 @@ string_concatenate( slist *tokens, fields *bibin, long nref, param *pm )
 				fprintf( stderr, "%s: Warning: String concentation should "
 					"be used in context of quotations marks in reference %ld\n", pm->progname, nref );
 			t = slist_str( tokens, i+1 );
-			if ( t->data[0]!='\"' && t->data[s->len-1]!='\"' )
+			if ( t->data[0]!='\"' && t->data[t->len-1]!='\"' )
 				fprintf( stderr, "%s: Warning: String concentation should "
 					"be used in context of quotations marks in reference %ld\n", pm->progname, nref );
 			if ( ( s->data[s->len-1]=='\"' && t->data[0]=='\"') || (s->data[s->len-1]=='}' && t->data[0]=='{') ) {
@@ -329,8 +335,8 @@ string_concatenate( slist *tokens, fields *bibin, long nref, param *pm )
 	return BIBL_OK;
 }
 
-static char *
-process_biblatexline( char *p, str *tag, str *data, uchar stripquotes, long nref, param *pm )
+static const char *
+process_biblatexline( const char *p, str *tag, str *data, uchar stripquotes, long nref, param *pm )
 {
 	int i, status;
 	slist tokens;
@@ -339,7 +345,12 @@ process_biblatexline( char *p, str *tag, str *data, uchar stripquotes, long nref
 	str_empty( data );
 
 	p = biblatex_tag( p, tag );
-	if ( tag->len==0 ) return p;
+	if ( str_is_empty( tag ) ) {
+		/* ...skip this line */
+		while ( *p && *p!='\n' && *p!='\r' ) p++;
+		while ( *p=='\n' || *p=='\r' ) p++;
+		return p;
+	}
 
 	slist_init( &tokens );
 
@@ -368,21 +379,24 @@ out:
 }
 
 static int
-process_cite( fields *bibin, char *p, char *filename, long nref, param *pm )
+process_cite( fields *bibin, const char *p, const char *filename, long nref, param *pm )
 {
 	int fstatus, status = BIBL_OK;
-	str tag, data;
-	strs_init( &tag, &data, NULL );
-	p = process_biblatextype( p, &data );
-	if ( str_has_value( &data ) ) {
-		fstatus = fields_add( bibin, "INTERNAL_TYPE", str_cstr( &data ), 0 );
-		if ( fstatus!=FIELDS_OK ) { status = BIBL_ERR_MEMERR; goto out; }
-	}
-	p = process_biblatexid ( p, &data );
-	if ( str_has_value( &data ) ) {
-		fstatus = fields_add( bibin, "REFNUM", str_cstr( &data ), 0 );
-		if ( fstatus!=FIELDS_OK ) { status = BIBL_ERR_MEMERR; goto out; }
-	}
+	str type, id, tag, data;
+
+	strs_init( &type, &id, &tag, &data, NULL );
+
+	p = process_biblatextype( p, &type );
+	p = process_biblatexid( p, &id );
+
+	if ( str_is_empty( &type ) || str_is_empty( &id ) ) goto out;
+
+	fstatus = fields_add( bibin, "INTERNAL_TYPE", str_cstr( &type ), 0 );
+	if ( fstatus!=FIELDS_OK ) { status = BIBL_ERR_MEMERR; goto out; }
+
+	fstatus = fields_add( bibin, "REFNUM", str_cstr( &id ), 0 );
+	if ( fstatus!=FIELDS_OK ) { status = BIBL_ERR_MEMERR; goto out; }
+
 	while ( *p ) {
 		p = process_biblatexline( p, &tag, &data, 1, nref, pm );
 		if ( !p ) { status = BIBL_ERR_MEMERR; goto out; }
@@ -394,7 +408,7 @@ process_cite( fields *bibin, char *p, char *filename, long nref, param *pm )
 		strs_empty( &tag, &data, NULL );
 	}
 out:
-	strs_free( &tag, &data, NULL );
+	strs_free( &type, &id, &tag, &data, NULL );
 	return status;
 }
 
@@ -410,26 +424,27 @@ out:
  *
  */
 static int
-process_string( char *p, long nref, param *pm )
+process_string( const char *p, long nref, param *pm )
 {
 	int n, status = BIBL_OK;
 	str s1, s2, *s;
 	strs_init( &s1, &s2, NULL );
 	while ( *p && *p!='{' && *p!='(' ) p++;
 	if ( *p=='{' || *p=='(' ) p++;
-	p = process_biblatexline( skip_ws( p ), &s1, &s2, 0, nref, pm );
+	(void) process_biblatexline( skip_ws( p ), &s1, &s2, 0, nref, pm );
 	if ( str_has_value( &s2 ) ) {
 		str_findreplace( &s2, "\\ ", " " );
 		if ( str_memerr( &s2 ) ) { status = BIBL_ERR_MEMERR; goto out; }
+	} else {
+		str_strcpyc( &s2, "" );
 	}
 	if ( str_has_value( &s1 ) ) {
 		n = slist_find( &find, &s1 );
 		if ( n==-1 ) {
-			s = slist_add( &find, &s1 );
-			if ( s==NULL ) { status = BIBL_ERR_MEMERR; goto out; }
-			if ( str_has_value( &s2 ) ) s = slist_add( &replace, &s2 );
-			else s = slist_addc( &replace, "" );
-			if ( s==NULL ) { status = BIBL_ERR_MEMERR; goto out; }
+			status = slist_add_ret( &find, &s1, BIBL_OK, BIBL_ERR_MEMERR );
+			if ( status!=BIBL_OK ) goto out;
+			status = slist_add_ret( &replace, &s2, BIBL_OK, BIBL_ERR_MEMERR );
+			if ( status!=BIBL_OK ) goto out;
 		} else {
 			if ( str_has_value( &s2 ) ) s = slist_set( &replace, n, &s2 );
 			else s = slist_setc( &replace, n, "" );
@@ -442,7 +457,7 @@ out:
 }
 
 static int
-biblatexin_processf( fields *bibin, char *data, char *filename, long nref, param *p )
+biblatexin_processf( fields *bibin, const char *data, const char *filename, long nref, param *p )
 {
 	if ( !strncasecmp( data, "@STRING", 7 ) ) {
 		process_string( data+7, nref, p );
@@ -457,163 +472,15 @@ biblatexin_processf( fields *bibin, char *data, char *filename, long nref, param
  PUBLIC: void biblatexin_cleanf()
 *****************************************************/
 
-static void
-biblatex_process_tilde( str *s )
-{
-	char *p, *q;
-	int n = 0;
-
-	p = q = s->data;
-	if ( !p ) return;
-	while ( *p ) {
-		if ( *p=='~' ) {
-			*q = ' ';
-		} else if ( *p=='\\' && *(p+1)=='~' ) {
-			n++;
-			p++;
-			*q = '~';
-		} else {
-			*q = *p;
-		}
-		p++;
-		q++;
-	}
-	*q = '\0';
-	s->len -= n;
-}
-
-static void
-biblatex_process_bracket( str *s )
-{
-	char *p, *q;
-	int n = 0;
-
-	p = q = s->data;
-	if ( !p ) return;
-	while ( *p ) {
-		if ( *p=='\\' && ( *(p+1)=='{' || *(p+1)=='}' ) ) {
-			n++;
-			p++;
-			*q = *p;
-			q++;
-		} else if ( *p=='{' || *p=='}' ) {
-			n++;
-		} else {
-			*q = *p;
-			q++;
-		}
-		p++;
-	}
-	*q = '\0';
-	s->len -= n;
-}
-
 static int
-biblatex_cleantoken( str *s )
+is_url_tag( str *tag )
 {
-	/* 'textcomp' annotations */
-	str_findreplace( s, "\\textit", "" );
-	str_findreplace( s, "\\textbf", "" );
-	str_findreplace( s, "\\textsl", "" );
-	str_findreplace( s, "\\textsc", "" );
-	str_findreplace( s, "\\textsf", "" );
-	str_findreplace( s, "\\texttt", "" );
-	str_findreplace( s, "\\textsubscript", "" );
-	str_findreplace( s, "\\textsuperscript", "" );
-	str_findreplace( s, "\\emph", "" );
-	str_findreplace( s, "\\url", "" );
-
-	/* Other text annotations */
-	str_findreplace( s, "\\it ", "" );
-	str_findreplace( s, "\\em ", "" );
-
-	str_findreplace( s, "\\%", "%" );
-	str_findreplace( s, "\\$", "$" );
-	while ( str_findreplace( s, "  ", " " ) ) {}
-
-	/* 'textcomp' annotations that we don't want to substitute on output*/
-	str_findreplace( s, "\\textdollar", "$" );
-	str_findreplace( s, "\\textunderscore", "_" );
-
-	biblatex_process_bracket( s );
-	biblatex_process_tilde( s );
-
-	if ( !str_memerr( s ) ) return BIBL_OK;
-	else return BIBL_ERR_MEMERR;
-}
-
-static int
-biblatex_split( slist *tokens, str *s )
-{
-	int i, n = s->len, nbrackets = 0, status = BIBL_OK;
-	str tok, *t;
-
-	str_init( &tok );
-
-	for ( i=0; i<n; ++i ) {
-		if ( s->data[i]=='{' && ( i==0 || s->data[i-1]!='\\' ) ) {
-			nbrackets++;
-			str_addchar( &tok, '{' );
-		} else if ( s->data[i]=='}' && ( i==0 || s->data[i-1]!='\\' ) ) {
-			nbrackets--;
-			str_addchar( &tok, '}' );
-		} else if ( !is_ws( s->data[i] ) || nbrackets ) {
-			str_addchar( &tok, s->data[i] );
-		} else if ( is_ws( s->data[i] ) ) {
-			if ( str_memerr( &tok ) ) { status = BIBL_ERR_MEMERR; goto out; }
-			if ( str_has_value( &tok ) ) {
-				t = slist_add( tokens, &tok );
-				if ( !t ) { status = BIBL_ERR_MEMERR; goto out; }
-			}
-			str_empty( &tok );
-		}
+	if ( str_has_value( tag ) ) {
+		if ( !strcasecmp( str_cstr( tag ), "url" ) ) return 1;
+		if ( !strcasecmp( str_cstr( tag ), "file" ) ) return 1;
+		if ( !strcasecmp( str_cstr( tag ), "doi" ) ) return 1;
 	}
-	if ( str_has_value( &tok ) ) {
-		if ( str_memerr( &tok ) ) { status = BIBL_ERR_MEMERR; goto out; }
-		t = slist_add( tokens, &tok );
-		if ( !t ) { status = BIBL_ERR_MEMERR; goto out; }
-	}
-
-	for ( i=0; i<tokens->n; ++i ) {
-		t = slist_str( tokens, i );
-		str_trimstartingws( t );
-		str_trimendingws( t );
-		if ( str_memerr( t ) ) { status = BIBL_ERR_MEMERR; goto out; }
-	}
-out:
-	str_free( &tok );
-	return status;
-}
-
-static int
-biblatexin_addtitleurl( fields *info, str *in )
-{
-	int fstatus, status = BIBL_OK;
-	char *p;
-	str s;
-
-	str_init( &s );
-	/* skip past "\href{" */
-	p = str_cpytodelim( &s, in->data + 6, "}", 1 );
-	if ( str_memerr( &s ) ) {
-		status = BIBL_ERR_MEMERR;
-		goto out;
-	}
-	fstatus = fields_add( info, "URL", s.data, 0 );
-	if ( fstatus!=FIELDS_OK ) {
-		status = BIBL_ERR_MEMERR;
-		goto out;
-	}
-
-	p = str_cpytodelim( &s, p, "", 0 );
-	if ( str_memerr( &s ) ) {
-		status = BIBL_ERR_MEMERR;
-		goto out;
-	}
-	str_swapstrings( &s, in );
-out:
-	str_free( &s );
-	return status;
+	return 0;
 }
 
 static int
@@ -640,69 +507,40 @@ is_name_tag( str *tag )
 }
 
 static int
-is_url_tag( str *tag )
+biblatexin_cleanvalue( str *tag, str *value, fields *bibin, param *p )
 {
-	if ( str_has_value( tag ) ) {
-		if ( !strcasecmp( str_cstr( tag ), "url" ) ) return 1;
-	}
-	return 0;
-}
+	int status = BIBL_OK;
+	str parsed;
+
+	if ( str_is_empty( value ) ) return BIBL_OK;
 
-static int
-biblatexin_cleandata( str *tag, str *s, fields *info, param *p )
-{
-	slist tokens;
-	str *tok;
-	int i, status = BIBL_OK;
-	if ( str_is_empty( s ) ) return status;
 	/* protect url from undergoing any parsing */
-	if ( is_url_tag( tag ) ) return status;
-	slist_init( &tokens );
-	biblatex_split( &tokens, s );
-	for ( i=0; i<tokens.n; ++i ) {
-		if ( !strncasecmp(slist_cstr( &tokens, i ), "\\href{", 6 )) {
-			status = biblatexin_addtitleurl( info, slist_str( &tokens, i ) );
-			if ( status!=BIBL_OK ) goto out;
-		}
-		if ( p && p->latexin && !is_name_tag( tag ) ) {
-			status = biblatex_cleantoken( slist_str( &tokens, i ) );
-			if ( status!=BIBL_OK ) goto out;
-		}
-	}
-	str_empty( s );
-	for ( i=0; i<tokens.n; ++i ) {
-		tok = slist_str( &tokens, i );
-		if ( i>0 ) str_addchar( s, ' ' );
-		str_strcat( s, tok );
-	}
-out:
-	slist_free( &tokens );
-	return status;
-}
+	if ( is_url_tag( tag ) ) return BIBL_OK;
 
-static long
-biblatexin_findref( bibl *bin, char *citekey )
-{
-	int n;
-	long i;
-	for ( i=0; i<bin->nrefs; ++i ) {
-		n = fields_find( bin->ref[i], "refnum", -1 );
-		if ( n==-1 ) continue;
-		if ( !strcmp( bin->ref[i]->data[n].data, citekey ) ) return i;
-	}
-	return -1;
+	/* delay names from undergoing any parsing */
+	if ( is_name_tag( tag ) ) return BIBL_OK;
+
+	str_init( &parsed );
+
+	status = latex_parse( value, &parsed );
+	if ( status!=BIBL_OK ) goto out;
+
+	str_strcpy( value, &parsed );
+	if ( str_memerr( value ) ) status = BIBL_ERR_MEMERR;
+
+out:
+	str_free( &parsed );
+	return status;
 }
 
 static void
 biblatexin_nocrossref( bibl *bin, long i, int n, param *p )
 {
-	int n1 = fields_find( bin->ref[i], "REFNUM", -1 );
+	int n1 = fields_find( bin->ref[i], "REFNUM", LEVEL_ANY );
 	if ( p->progname ) fprintf( stderr, "%s: ", p->progname );
-	fprintf( stderr, "Cannot find cross-reference '%s'", 
-			bin->ref[i]->data[n].data);
-	if ( n1!=-1 )
-		fprintf( stderr, " for reference '%s'\n", 
-				bin->ref[i]->data[n1].data );
+	fprintf( stderr, "Cannot find cross-reference '%s'", (char*)fields_value( bin->ref[i], n, FIELDS_CHRP_NOUSE ) );
+	if ( n1!=FIELDS_NOTFOUND )
+		fprintf( stderr, " for reference '%s'", (char*)fields_value( bin->ref[i], n1, FIELDS_CHRP_NOUSE ) );
 	fprintf( stderr, "\n" );
 }
 
@@ -711,7 +549,7 @@ biblatexin_crossref_oneref( fields *ref, fields *cross )
 {
 	int j, nl, ntype, fstatus;
 	char *type, *nt, *nd;
-	ntype = fields_find( ref, "INTERNAL_TYPE", -1 );
+	ntype = fields_find( ref, "INTERNAL_TYPE", LEVEL_ANY );
 	type = ( char * ) fields_value( ref, ntype, FIELDS_CHRP_NOUSE );
 	for ( j=0; j<cross->n; ++j ) {
 		nt = ( char * ) fields_tag( cross, j, FIELDS_CHRP_NOUSE );
@@ -736,12 +574,12 @@ biblatexin_crossref( bibl *bin, param *p )
 	int n, ncross, status = BIBL_OK;
 	fields *ref, *cross;
 	long i;
-        for ( i=0; i<bin->nrefs; ++i ) {
+        for ( i=0; i<bin->n; ++i ) {
 		ref = bin->ref[i];
-		n = fields_find( ref, "CROSSREF", -1 );
-		if ( n==-1 ) continue;
-		fields_setused( ref, n );
-		ncross = biblatexin_findref(bin, (char*)fields_value(ref,n, FIELDS_CHRP_NOUSE));
+		n = fields_find( ref, "CROSSREF", LEVEL_ANY );
+		if ( n==FIELDS_NOTFOUND ) continue;
+		fields_set_used( ref, n );
+		ncross = bibl_findref( bin, (char*) fields_value(ref,n, FIELDS_CHRP_NOUSE) );
 		if ( ncross==-1 ) {
 			biblatexin_nocrossref( bin, i, n, p );
 			continue;
@@ -762,7 +600,7 @@ biblatexin_cleanref( fields *bibin, param *p )
 	for ( i=0; i<n; ++i ) {
 		t = fields_tag( bibin, i, FIELDS_STRP_NOUSE );
 		d = fields_value( bibin, i, FIELDS_STRP_NOUSE );
-		status = biblatexin_cleandata( t, d, bibin, p );
+		status = biblatexin_cleanvalue( t, d, bibin, p );
 		if ( status!=BIBL_OK ) return status;
 		if ( !strsearch( str_cstr( t ), "AUTHORS" ) ) {
 			str_findreplace( d, "\n", " " );
@@ -783,7 +621,7 @@ biblatexin_cleanf( bibl *bin, param *p )
 {
 	int status;
 	long i;
-        for ( i=0; i<bin->nrefs; ++i ) {
+        for ( i=0; i<bin->n; ++i ) {
 		status = biblatexin_cleanref( bin->ref[i], p );
 		if ( status!=BIBL_OK ) return status;
 	}
@@ -796,15 +634,15 @@ biblatexin_cleanf( bibl *bin, param *p )
 *****************************************************/
 
 static int
-biblatexin_typef( fields *bibin, char *filename, int nrefs, param *p )
+biblatexin_typef( fields *bibin, const char *filename, int nrefs, param *p )
 {
 	int ntypename, nrefname, is_default;
 	char *refname = "", *typename="";
 
 	ntypename = fields_find( bibin, "INTERNAL_TYPE", LEVEL_MAIN );
 	nrefname  = fields_find( bibin, "REFNUM",        LEVEL_MAIN );
-	if ( nrefname!=-1 )  refname  = fields_value( bibin, nrefname,  FIELDS_CHRP_NOUSE );
-        if ( ntypename!=-1 ) typename = fields_value( bibin, ntypename, FIELDS_CHRP_NOUSE );
+	if ( nrefname!=FIELDS_NOTFOUND )  refname  = fields_value( bibin, nrefname,  FIELDS_CHRP_NOUSE );
+        if ( ntypename!=FIELDS_NOTFOUND ) typename = fields_value( bibin, ntypename, FIELDS_CHRP_NOUSE );
 
 	return get_reftype( typename, nrefs, p->progname, p->all, p->nall, refname, &is_default, REFTYPE_CHATTY );
 }
@@ -850,7 +688,7 @@ get_title_elements( fields *bibin, int currlevel, int reftype, variants *all, in
 		if ( process != TITLE ) continue;
 		if ( level != currlevel ) continue;
 
-		fields_setused( bibin, i );
+		fields_set_used( bibin, i );
 
 		if ( !strcasecmp( newtag, "TITLE" ) ) {
 			if ( str_has_value( ttl ) ) str_addchar( ttl, ' ' );
@@ -918,7 +756,7 @@ process_combined_title( fields *info, str *ttl, str *subttl, str *ttladdon, int
 		goto out;
 	}
 
-	fstatus = fields_add( info, "TITLE", combined.data, currlevel );
+	fstatus = fields_add( info, "TITLE", str_cstr( &combined ), currlevel );
 	if ( fstatus==FIELDS_OK ) status = BIBL_ERR_MEMERR;
 out:
 	str_free( &combined );
@@ -999,6 +837,7 @@ biblatex_names( fields *info, char *tag, str *data, int level, slist *asis, slis
 {
 	int begin, end, ok, n, etal, i, match, status = BIBL_OK;
 	slist tokens;
+	str parsed, *s;
 
 	/* If we match the asis or corps list add and bail. */
 	status = biblatex_matches_list( info, tag, ":ASIS", data, level, asis, &match );
@@ -1007,10 +846,17 @@ biblatex_names( fields *info, char *tag, str *data, int level, slist *asis, slis
 	if ( match==1 || status!=BIBL_OK ) return status;
 
 	slist_init( &tokens );
+	str_init( &parsed );
 
-	biblatex_split( &tokens, data );
-	for ( i=0; i<tokens.n; ++i )
-		biblatex_cleantoken( slist_str( &tokens, i ) );
+	status = latex_tokenize( &tokens, data );
+	if ( status!=BIBL_OK ) goto out;
+
+	for ( i=0; i<tokens.n; ++i ) {
+		status = latex_parse( slist_str( &tokens, i ), &parsed );
+		if ( status!=BIBL_OK ) goto out;
+		s = slist_set( &tokens, i, &parsed );
+		if ( !s ) { status=BIBL_ERR_MEMERR; goto out; }
+	}
 
 	etal = name_findetal( &tokens );
 
@@ -1045,6 +891,7 @@ biblatex_names( fields *info, char *tag, str *data, int level, slist *asis, slis
 	}
 
 out:
+	str_free( &parsed );
 	slist_free( &tokens );
 	return status;
 }
@@ -1055,14 +902,14 @@ biblatexin_bltsubtype( fields *bibin, int n, str *intag, str *invalue, int level
 	int fstatus1, fstatus2;
 
 	if ( !strcasecmp( str_cstr( invalue ), "magazine" ) ) {
-		fstatus1 = fields_add( bibout, "NGENRE", "magazine article", LEVEL_MAIN );
-		fstatus2 = fields_add( bibout, "NGENRE", "magazine",         LEVEL_HOST );
+		fstatus1 = fields_add( bibout, "GENRE:BIBUTILS", "magazine article", LEVEL_MAIN );
+		fstatus2 = fields_add( bibout, "GENRE:BIBUTILS", "magazine",         LEVEL_HOST );
 		if ( fstatus1!=FIELDS_OK || fstatus2!=FIELDS_OK ) return BIBL_ERR_MEMERR;
 	}
 
 	else if ( !strcasecmp( str_cstr( invalue ), "newspaper" ) ) {
-		fstatus1 = fields_add( bibout, "NGENRE", "newspaper article", LEVEL_MAIN );
-		fstatus2 = fields_add( bibout, "GENRE",  "newspaper",         LEVEL_HOST );
+		fstatus1 = fields_add( bibout, "GENRE:BIBUTILS", "newspaper article", LEVEL_MAIN );
+		fstatus2 = fields_add( bibout, "GENRE:MARC",     "newspaper",         LEVEL_HOST );
 		if ( fstatus1!=FIELDS_OK || fstatus2!=FIELDS_OK ) return BIBL_ERR_MEMERR;
 	}
 
@@ -1074,7 +921,7 @@ static int
 biblatexin_bltschool( fields *bibin, int n, str *intag, str *invalue, int level, param *pm, char *outtag, fields *bibout )
 {
 	int fstatus;
-	if ( fields_find( bibin, "institution", LEVEL_ANY ) != -1 )
+	if ( fields_find( bibin, "institution", LEVEL_ANY ) != FIELDS_NOTFOUND )
 		return BIBL_OK;
 	else {
 		fstatus = fields_add( bibout, outtag, str_cstr( invalue ), level );
@@ -1090,15 +937,15 @@ biblatexin_bltthesistype( fields *bibin, int n, str *intag, str *invalue, int le
 	int fstatus = FIELDS_OK;
 	/* type in the @thesis is used to distinguish Ph.D. and Master's thesis */
 	if ( !strncasecmp( p, "phdthesis", 9 ) ) {
-		fstatus = fields_replace_or_add( bibout, "NGENRE", "Ph.D. thesis", level );
+		fstatus = fields_replace_or_add( bibout, "GENRE:BIBUTILS", "Ph.D. thesis", level );
 	} else if ( !strncasecmp( p, "mastersthesis", 13 ) || !strncasecmp( p, "masterthesis", 12 ) ) {
-		fstatus = fields_replace_or_add( bibout, "NGENRE", "Masters thesis", level );
+		fstatus = fields_replace_or_add( bibout, "GENRE:BIBUTILS", "Masters thesis", level );
 	} else if ( !strncasecmp( p, "mathesis", 8 ) ) {
-		fstatus = fields_replace_or_add( bibout, "NGENRE", "Masters thesis", level );
+		fstatus = fields_replace_or_add( bibout, "GENRE:BIBUTILS", "Masters thesis", level );
 	} else if ( !strncasecmp( p, "diploma", 7 ) ) {
-		fstatus = fields_replace_or_add( bibout, "NGENRE", "Diploma thesis", level );
+		fstatus = fields_replace_or_add( bibout, "GENRE:BIBUTILS", "Diploma thesis", level );
 	} else if ( !strncasecmp( p, "habilitation", 12 ) ) {
-		fstatus = fields_replace_or_add( bibout, "NGENRE", "Habilitation thesis", level );
+		fstatus = fields_replace_or_add( bibout, "GENRE:BIBUTILS", "Habilitation thesis", level );
 	}
 	if ( fstatus==FIELDS_OK ) return BIBL_OK;
 	else return BIBL_ERR_MEMERR;
@@ -1110,11 +957,11 @@ biblatexin_bteprint( fields *bibin, int n, str *intag, str *invalue, int level,
 	int neprint, netype, fstatus;
 	char *eprint = NULL, *etype = NULL;
 
-	neprint = fields_find( bibin, "eprint", -1 );
-	netype  = fields_find( bibin, "eprinttype", -1 );
+	neprint = fields_find( bibin, "eprint",     LEVEL_ANY );
+	netype  = fields_find( bibin, "eprinttype", LEVEL_ANY );
 
-	if ( neprint!=-1 ) eprint = bibin->data[neprint].data;
-	if ( netype!=-1 )  etype =  bibin->data[netype].data;
+	if ( neprint!=FIELDS_NOTFOUND ) eprint = fields_value( bibin, neprint, FIELDS_CHRP );
+	if ( netype!=FIELDS_NOTFOUND )  etype =  fields_value( bibin, netype,  FIELDS_CHRP );
 
 	if ( eprint && etype ) {
 		if ( !strncasecmp( etype, "arxiv", 5 ) ) {
@@ -1135,16 +982,16 @@ biblatexin_bteprint( fields *bibin, int n, str *intag, str *invalue, int level,
 			fstatus = fields_add( bibout, "EPRINTTYPE", etype, level );
 			if ( fstatus!=FIELDS_OK ) return BIBL_ERR_MEMERR;
 		}
-		fields_setused( bibin, neprint );
-		fields_setused( bibin, netype );
+		fields_set_used( bibin, neprint );
+		fields_set_used( bibin, netype );
 	} else if ( eprint ) {
 		fstatus = fields_add( bibout, "EPRINT", eprint, level );
 		if ( fstatus!=FIELDS_OK ) return BIBL_ERR_MEMERR;
-		fields_setused( bibin, neprint );
+		fields_set_used( bibin, neprint );
 	} else if ( etype ) {
 		fstatus = fields_add( bibout, "EPRINTTYPE", etype, level );
 		if ( fstatus!=FIELDS_OK ) return BIBL_ERR_MEMERR;
-		fields_setused( bibin, netype );
+		fields_set_used( bibin, netype );
 	}
 	return BIBL_OK;
 }
@@ -1152,7 +999,7 @@ biblatexin_bteprint( fields *bibin, int n, str *intag, str *invalue, int level,
 static int
 biblatexin_btgenre( fields *bibin, int n, str *intag, str *invalue, int level, param *pm, char *outtag, fields *bibout )
 {
-	if ( fields_add( bibout, "NGENRE", str_cstr( invalue ), level ) == FIELDS_OK ) return BIBL_OK;
+	if ( fields_add( bibout, "GENRE:BIBUTILS", str_cstr( invalue ), level ) == FIELDS_OK ) return BIBL_OK;
 	else return BIBL_ERR_MEMERR;
 }
 
@@ -1174,9 +1021,9 @@ biblatexin_howpublished( fields *bibin, int n, str *intag, str *invalue, int lev
 	int fstatus;
 
         if ( !strncasecmp( str_cstr( invalue ), "Diplom", 6 ) )
-                fstatus = fields_replace_or_add( bibout, "NGENRE", "Diploma thesis", level );
+                fstatus = fields_replace_or_add( bibout, "GENRE:BIBUTILS", "Diploma thesis", level );
         else if ( !strncasecmp( str_cstr( invalue ), "Habilitation", 13 ) )
-                fstatus = fields_replace_or_add( bibout, "NGENRE", "Habilitation thesis", level );
+                fstatus = fields_replace_or_add( bibout, "GENRE:BIBUTILS", "Habilitation thesis", level );
         else
 		fstatus = fields_add( bibout, "PUBLISHER", str_cstr( invalue ), level );
 
@@ -1216,7 +1063,7 @@ biblatexin_blteditor( fields *bibin, int m, str *intag, str *invalue, int level,
 	for ( i=1; i<neditors; ++i )
 		if ( !strcasecmp( intag->data, editor_fields[i] ) ) n = i;
 	ntype = fields_find( bibin, editor_types[n], LEVEL_ANY );
-	if ( ntype!=-1 ) {
+	if ( ntype!=FIELDS_NOTFOUND ) {
 		type = fields_value( bibin, ntype, FIELDS_CHRP_NOUSE );
 		if ( !strcasecmp( type, "collaborator" ) )  usetag = "COLLABORATOR";
 		else if ( !strcasecmp( type, "compiler" ) ) usetag = "COMPILER";
@@ -1255,7 +1102,7 @@ biblatexin_convertf( fields *bibin, fields *bibout, int reftype, param *p )
 		[ BLT_EDITOR      ] = biblatexin_blteditor,
 		[ HOWPUBLISHED    ] = biblatexin_howpublished,
 		[ URL             ] = generic_url,
-		[ BT_GENRE        ] = biblatexin_btgenre,
+		[ GENRE           ] = biblatexin_btgenre,
 		[ BT_EPRINT       ] = biblatexin_bteprint,
 		[ BLT_THESIS_TYPE ] = biblatexin_bltthesistype,
 		[ BLT_SCHOOL      ] = biblatexin_bltschool,
@@ -1288,7 +1135,7 @@ biblatexin_convertf( fields *bibin, fields *bibout, int reftype, param *p )
 		if ( status!=BIBL_OK ) return status;
 
 		if ( convertfns[ process ] != generic_null )
-			fields_setused( bibin, i );
+			fields_set_used( bibin, i );
 
 	}
 
