@@ -1,7 +1,7 @@
 /*
  * medin.c
  *
- * Copyright (c) Chris Putnam 2004-2017
+ * Copyright (c) Chris Putnam 2004-2020
  *
  * Source code released under the GPL version 2
  *
@@ -19,40 +19,45 @@
 #include "bibformats.h"
 
 static int medin_readf( FILE *fp, char *buf, int bufsize, int *bufpos, str *line, str *reference, int *fcharset );
-static int medin_processf( fields *medin, char *data, char *filename, long nref, param *p );
+static int medin_processf( fields *medin, const char *data, const char *filename, long nref, param *p );
 
 
 /*****************************************************
  PUBLIC: void medin_initparams()
 *****************************************************/
-void
-medin_initparams( param *p, const char *progname )
+int
+medin_initparams( param *pm, const char *progname )
 {
-	p->readformat       = BIBL_MEDLINEIN;
-	p->charsetin        = BIBL_CHARSET_UNICODE;
-	p->charsetin_src    = BIBL_SRC_DEFAULT;
-	p->latexin          = 0;
-	p->xmlin            = 1;
-	p->utf8in           = 1;
-	p->nosplittitle     = 0;
-	p->verbose          = 0;
-	p->addcount         = 0;
-	p->output_raw       = BIBL_RAW_WITHMAKEREFID |
+	pm->readformat       = BIBL_MEDLINEIN;
+	pm->charsetin        = BIBL_CHARSET_UNICODE;
+	pm->charsetin_src    = BIBL_SRC_DEFAULT;
+	pm->latexin          = 0;
+	pm->xmlin            = 1;
+	pm->utf8in           = 1;
+	pm->nosplittitle     = 0;
+	pm->verbose          = 0;
+	pm->addcount         = 0;
+	pm->output_raw       = BIBL_RAW_WITHMAKEREFID |
 	                      BIBL_RAW_WITHCHARCONVERT;
 
-	p->readf    = medin_readf;
-	p->processf = medin_processf;
-	p->cleanf   = NULL;
-	p->typef    = NULL;
-	p->convertf = NULL;
-	p->all      = NULL;
-	p->nall     = 0;
+	pm->readf    = medin_readf;
+	pm->processf = medin_processf;
+	pm->cleanf   = NULL;
+	pm->typef    = NULL;
+	pm->convertf = NULL;
+	pm->all      = NULL;
+	pm->nall     = 0;
 
-	slist_init( &(p->asis) );
-	slist_init( &(p->corps) );
+	slist_init( &(pm->asis) );
+	slist_init( &(pm->corps) );
 
-	if ( !progname ) p->progname = NULL;
-	else p->progname = strdup( progname );
+	if ( !progname ) pm->progname = NULL;
+	else {
+		pm->progname = strdup( progname );
+		if ( !pm->progname ) return BIBL_ERR_MEMERR;
+	}
+
+	return BIBL_OK;
 }
 
 /*****************************************************
@@ -73,7 +78,7 @@ medin_findstartwrapper( char *buf, int *ntype )
 	char *startptr=NULL;
 	int i;
 	for ( i=0; i<nwrapper && startptr==NULL; ++i ) {
-		startptr = xml_findstart( buf, wrapper[ i ] );
+		startptr = xml_find_start( buf, wrapper[ i ] );
 		if ( startptr && *ntype==-1 ) *ntype = i;
 	}
 	return startptr;
@@ -82,7 +87,7 @@ medin_findstartwrapper( char *buf, int *ntype )
 static char *
 medin_findendwrapper( char *buf, int ntype )
 {
-	char *endptr = xml_findend( buf, wrapper[ ntype ] );
+	char *endptr = xml_find_end( buf, wrapper[ ntype ] );
 	return endptr;
 }
 
@@ -136,17 +141,17 @@ medin_doconvert( xml *node, fields *info, xml_convert *c, int nc, int *found )
 	int i, fstatus;
 	char *d;
 	*found = 0;
-	if ( !xml_hasvalue( node ) ) return BIBL_OK;
-	d = xml_value( node );
+	if ( !xml_has_value( node ) ) return BIBL_OK;
+	d = xml_value_cstr( node );
 	for ( i=0; i<nc && *found==0; ++i ) {
 		if ( c[i].a==NULL ) {
-			if ( xml_tagexact( node, c[i].in ) ) {
+			if ( xml_tag_matches( node, c[i].in ) ) {
 				*found = 1;
 				fstatus = fields_add( info, c[i].out, d, c[i].level );
 				if ( fstatus!=FIELDS_OK ) return BIBL_ERR_MEMERR;
 			}
 		} else {
-			if ( xml_tag_attrib( node, c[i].in, c[i].a, c[i].aval)){
+			if ( xml_tag_has_attribute( node, c[i].in, c[i].a, c[i].aval ) ) {
 				*found = 1;
 				fstatus = fields_add( info, c[i].out, d, c[i].level );
 				if ( fstatus!=FIELDS_OK ) return BIBL_ERR_MEMERR;
@@ -163,8 +168,8 @@ static int
 medin_articletitle( xml *node, fields *info )
 {
 	int fstatus, status = BIBL_OK;
-	if ( xml_hasvalue( node ) ) {
-		fstatus = fields_add( info, "TITLE", xml_value( node ), 0 );
+	if ( xml_has_value( node ) ) {
+		fstatus = fields_add( info, "TITLE", xml_value_cstr( node ), LEVEL_MAIN );
 		if ( fstatus!=FIELDS_OK ) status = BIBL_ERR_MEMERR;
 	}
 	return status;
@@ -172,7 +177,7 @@ medin_articletitle( xml *node, fields *info )
 
 /*            <MedlineDate>2003 Jan-Feb</MedlineDate> */
 static int
-medin_medlinedate( fields *info, char *p, int level )
+medin_medlinedate( fields *info, const char *p, int level )
 {
 	int fstatus;
 	str tmp;
@@ -196,7 +201,7 @@ medin_medlinedate( fields *info, char *p, int level )
 		if ( fstatus!=FIELDS_OK ) return BIBL_ERR_MEMERR;
 	}
 
-	p = str_cpytodelim( &tmp, skip_ws( p ), " \t\n\r", 0 );
+	(void) str_cpytodelim( &tmp, skip_ws( p ), " \t\n\r", 0 );
 	if ( str_memerr( &tmp ) ) return BIBL_ERR_MEMERR;
 
 	if ( str_has_value( &tmp ) ) {
@@ -216,7 +221,7 @@ medin_language( xml *node, fields *info, int level )
 {
 	char *code, *language;
 	int fstatus;
-	code = xml_value( node );
+	code = xml_value_cstr( node );
 	if ( !code ) return BIBL_OK;
 	language = iso639_2_from_code( code );
 	if ( language )
@@ -269,16 +274,16 @@ medin_journal1( xml *node, fields *info )
 		{ "Day",             NULL, NULL, "PARTDATE:DAY",   1 },
 	};
 	int nc = sizeof( c ) / sizeof( c[0] ), status, found;
-	if ( xml_hasvalue( node ) ) {
+	if ( xml_has_value( node ) ) {
 		status = medin_doconvert( node, info, c, nc, &found );
 		if ( status!=BIBL_OK ) return status;
 		if ( !found ) {
-			if ( xml_tagexact( node, "MedlineDate" ) ) {
-				status = medin_medlinedate( info, xml_value( node ), 1 );
+			if ( xml_tag_matches( node, "MedlineDate" ) ) {
+				status = medin_medlinedate( info, xml_value_cstr( node ), 1 );
 				if ( status!=BIBL_OK ) return status;
 			}
-			if ( xml_tagexact( node, "Language" ) ) {
-				status = medin_language( node, info, 1 );
+			if ( xml_tag_matches( node, "Language" ) ) {
+				status = medin_language( node, info, LEVEL_HOST );
 				if ( status!=BIBL_OK ) return status;
 			}
 		}
@@ -303,16 +308,16 @@ medin_pagination( xml *node, fields *info )
 {
 	int i, fstatus, status;
 	str sp, ep;
-	char *p, *pp;
-	if ( xml_tagexact( node, "MedlinePgn" ) && node->value ) {
+	const char *p, *pp;
+	if ( xml_tag_matches( node, "MedlinePgn" ) && node->value.len ) {
 		strs_init( &sp, &ep, NULL );
-		p = str_cpytodelim( &sp, xml_value( node ), "-", 1 );
+		p = str_cpytodelim( &sp, xml_value_cstr( node ), "-", 1 );
 		if ( str_memerr( &sp ) ) return BIBL_ERR_MEMERR;
 		if ( str_has_value( &sp ) ) {
-			fstatus = fields_add( info, "PAGES:START", str_cstr( &sp ), 1 );
+			fstatus = fields_add( info, "PAGES:START", str_cstr( &sp ), LEVEL_HOST );
 			if ( fstatus!=FIELDS_OK ) return BIBL_ERR_MEMERR;
 		}
-		p = str_cpytodelim( &ep, p, "", 0 );
+		(void) str_cpytodelim( &ep, p, "", 0 );
 		if ( str_memerr( &ep ) ) return BIBL_ERR_MEMERR;
 		if ( str_has_value( &ep ) ) {
 			if ( sp.len > ep.len ) {
@@ -320,7 +325,7 @@ medin_pagination( xml *node, fields *info )
 					sp.data[i] = ep.data[i-sp.len+ep.len];
 				pp = sp.data;
 			} else  pp = ep.data;
-			fstatus = fields_add( info, "PAGES:STOP", pp, 1 );
+			fstatus = fields_add( info, "PAGES:STOP", pp, LEVEL_HOST );
 			if ( fstatus!=FIELDS_OK ) return BIBL_ERR_MEMERR;
 		}
 		strs_free( &sp, &ep, NULL );
@@ -344,8 +349,8 @@ static int
 medin_abstract( xml *node, fields *info )
 {
 	int fstatus;
-	if ( xml_tagwithvalue( node, "AbstractText" ) ) {
-		fstatus = fields_add( info, "ABSTRACT", xml_value( node ), 0 );
+	if ( xml_tag_matches_has_value( node, "AbstractText" ) ) {
+		fstatus = fields_add( info, "ABSTRACT", xml_value_cstr( node ), LEVEL_MAIN );
 		if ( fstatus!=FIELDS_OK ) return BIBL_ERR_MEMERR;
 	} else if ( node->next ) return medin_abstract( node->next, info );
 	return BIBL_OK;
@@ -367,22 +372,22 @@ static int
 medin_author( xml *node, str *name )
 {
 	char *p;
-	if ( xml_tagexact( node, "LastName" ) ) {
+	if ( xml_tag_matches( node, "LastName" ) ) {
 		if ( str_has_value( name ) ) {
 			str_prepend( name, "|" );
-			str_prepend( name, xml_value( node ) );
+			str_prepend( name, xml_value_cstr( node ) );
 		}
-		else str_strcatc( name, xml_value( node ) );
-	} else if ( xml_tagexact( node, "ForeName" ) || 
-	            xml_tagexact( node, "FirstName" ) ) {
-		p = xml_value( node );
+		else str_strcat( name, xml_value( node ) );
+	} else if ( xml_tag_matches( node, "ForeName" ) ||
+	            xml_tag_matches( node, "FirstName" ) ) {
+		p = xml_value_cstr( node );
 		while ( p && *p ) {
 			if ( str_has_value( name ) ) str_addchar( name, '|' );
-			while ( *p && *p==' ' ) p++;
+			while ( *p==' ' ) p++;
 			while ( *p && *p!=' ' ) str_addchar( name, *p++ );
 		}
-	} else if ( xml_tagexact( node, "Initials" ) && !strchr( name->data, '|' )) {
-		p = xml_value( node );
+	} else if ( xml_tag_matches( node, "Initials" ) && !strchr( name->data, '|' )) {
+		p = xml_value_cstr( node );
 		while ( p && *p ) {
 			if ( str_has_value( name ) ) str_addchar( name, '|' );
 			if ( !is_ws(*p) ) str_addchar( name, *p++ );
@@ -395,8 +400,8 @@ medin_author( xml *node, str *name )
 static int
 medin_corpauthor( xml *node, str *name )
 {
-	if ( xml_tagexact( node, "CollectiveName" ) ) {
-		str_strcpyc( name, xml_value( node ) );
+	if ( xml_tag_matches( node, "CollectiveName" ) ) {
+		str_strcpy( name, xml_value( node ) );
 	} else if ( node->next ) medin_corpauthor( node->next, name );
 	return BIBL_OK;
 }
@@ -404,28 +409,43 @@ medin_corpauthor( xml *node, str *name )
 static int
 medin_authorlist( xml *node, fields *info )
 {
-	int fstatus, status;
-	str name;
+	int fstatus, status = BIBL_OK;
 	char *tag;
+	str name;
+
 	str_init( &name );
+
 	node = node->down;
 	while ( node ) {
-		if ( xml_tagexact( node, "Author" ) && node->down ) {
+
+		if ( xml_tag_matches( node, "Author" ) && node->down ) {
+
 			status = medin_author( node->down, &name );
+			if ( status!=BIBL_OK ) goto out;
+
 			tag = "AUTHOR";
 			if ( str_is_empty( &name ) ) {
 				status = medin_corpauthor( node->down, &name );
 				tag = "AUTHOR:CORP";
 			}
-			if ( str_memerr( &name ) || status!=BIBL_OK ) return BIBL_ERR_MEMERR;
+			if ( str_memerr( &name ) || status!=BIBL_OK ) {
+				status = BIBL_ERR_MEMERR;
+				goto out;
+			}
 			if ( str_has_value( &name ) ) {
-				fstatus = fields_add(info,tag,name.data,0);
-				if ( fstatus!=FIELDS_OK ) return BIBL_ERR_MEMERR;
+				fstatus = fields_add( info, tag, str_cstr( &name ), LEVEL_MAIN );
+				if ( fstatus!=FIELDS_OK ) {
+					status = BIBL_ERR_MEMERR;
+					goto out;
+				}
 			}
 			str_empty( &name );
 		}
+
 		node = node->next;
 	}
+
+out:
 	str_free( &name );
 	return BIBL_OK;
 }
@@ -446,8 +466,8 @@ static int
 medin_journal2( xml *node, fields *info )
 {
 	int fstatus, status = BIBL_OK;
-	if ( xml_tagwithvalue( node, "MedlineTA" ) && fields_find( info, "TITLE", 1 )==-1 ) {
-		fstatus = fields_add( info, "TITLE", xml_value( node ), 1 );
+	if ( xml_tag_matches_has_value( node, "MedlineTA" ) && fields_find( info, "TITLE", LEVEL_HOST )==FIELDS_NOTFOUND ) {
+		fstatus = fields_add( info, "TITLE", xml_value_cstr( node ), 1 );
 		if ( fstatus!=FIELDS_OK ) return BIBL_ERR_MEMERR;
 	}
 	if ( node->down ) {
@@ -472,8 +492,8 @@ static int
 medin_meshheading( xml *node, fields *info )
 {
 	int fstatus, status = BIBL_OK;
-	if ( xml_tagwithvalue( node, "DescriptorName" ) ) {
-		fstatus = fields_add( info, "KEYWORD", xml_value( node ), 0 );
+	if ( xml_tag_matches_has_value( node, "DescriptorName" ) ) {
+		fstatus = fields_add( info, "KEYWORD", xml_value_cstr( node ), 0 );
 		if ( fstatus!=FIELDS_OK ) return BIBL_ERR_MEMERR;
 	}
 	if ( node->next ) status = medin_meshheading( node->next, info );
@@ -484,7 +504,7 @@ static int
 medin_meshheadinglist( xml *node, fields *info )
 {
 	int status = BIBL_OK;
-	if ( xml_tagexact( node, "MeshHeading" ) && node->down ) {
+	if ( xml_tag_matches( node, "MeshHeading" ) && node->down ) {
 		status = medin_meshheading( node->down, info );
 		if ( status!=BIBL_OK ) return status;
 	}
@@ -522,7 +542,7 @@ medin_pubmeddata( xml *node, fields *info )
 		if ( status!=BIBL_OK ) return status;
 	}
 	if ( node->down ) {
-		medin_pubmeddata( node->down, info );
+		status = medin_pubmeddata( node->down, info );
 		if ( status!=BIBL_OK ) return status;
 	}
 	return BIBL_OK;
@@ -532,20 +552,20 @@ static int
 medin_article( xml *node, fields *info )
 {
 	int fstatus, status = BIBL_OK;
-	if ( xml_tagexact( node, "Journal" ) )
+	if ( xml_tag_matches( node, "Journal" ) )
 		status = medin_journal1( node, info );
-	else if ( xml_tagexact( node, "ArticleTitle" ) )
+	else if ( xml_tag_matches( node, "ArticleTitle" ) )
 		status = medin_articletitle( node, info );
-	else if ( xml_tagexact( node, "Pagination" ) && node->down )
+	else if ( xml_tag_matches( node, "Pagination" ) && node->down )
 		status = medin_pagination( node->down, info );
-	else if ( xml_tagexact( node, "Abstract" ) && node->down )
+	else if ( xml_tag_matches( node, "Abstract" ) && node->down )
 		status = medin_abstract( node->down, info );
-	else if ( xml_tagexact( node, "AuthorList" ) )
+	else if ( xml_tag_matches( node, "AuthorList" ) )
 		status = medin_authorlist( node, info );
-	else if ( xml_tagexact( node, "Language" ) )
-		status = medin_language( node, info, 0 );
-	else if ( xml_tagexact( node, "Affiliation" ) ) {
-		fstatus = fields_add( info, "ADDRESS", xml_value( node ), 0 );
+	else if ( xml_tag_matches( node, "Language" ) )
+		status = medin_language( node, info, LEVEL_MAIN );
+	else if ( xml_tag_matches( node, "Affiliation" ) ) {
+		fstatus = fields_add( info, "ADDRESS", xml_value_cstr( node ), LEVEL_MAIN );
 		if ( fstatus!=FIELDS_OK ) status = BIBL_ERR_MEMERR;
 	}
 	if ( status!=BIBL_OK ) return status;
@@ -557,16 +577,16 @@ static int
 medin_medlinecitation( xml *node, fields *info )
 {
 	int fstatus, status = BIBL_OK;
-	if ( xml_tagexact( node, "PMID" ) && node->value->data ) {
-		fstatus = fields_add( info, "PMID", node->value->data, 0 );
+	if ( xml_tag_matches_has_value( node, "PMID" ) ) {
+		fstatus = fields_add( info, "PMID", xml_value_cstr( node ), LEVEL_MAIN );
 		if ( fstatus!=FIELDS_OK ) return BIBL_ERR_MEMERR;
 	}
 	if ( node->down ) {
-		if ( xml_tagexact( node, "Article" ) ) {
+		if ( xml_tag_matches( node, "Article" ) ) {
 			status = medin_article( node->down, info );
-		} else if ( xml_tagexact( node, "MedlineJournalInfo" ) ) {
+		} else if ( xml_tag_matches( node, "MedlineJournalInfo" ) ) {
 			status = medin_journal2( node->down, info );
-		} else if ( xml_tagexact( node, "MeshHeadingList" ) )
+		} else if ( xml_tag_matches( node, "MeshHeadingList" ) )
 			status = medin_meshheadinglist( node->down, info );
 		if ( status!=BIBL_OK ) return status;
 	}
@@ -579,9 +599,9 @@ medin_pubmedarticle( xml *node, fields *info )
 {
 	int status = BIBL_OK;
 	if ( node->down ) {
-		if ( xml_tagexact( node, "MedlineCitation" ) )
+		if ( xml_tag_matches( node, "MedlineCitation" ) )
 			status = medin_medlinecitation( node->down, info );
-		else if ( xml_tagexact( node, "PubmedData" ) )
+		else if ( xml_tag_matches( node, "PubmedData" ) )
 			status = medin_pubmeddata( node->down, info );
 		if ( status!=BIBL_OK ) return status;
 	}
@@ -594,9 +614,9 @@ medin_assembleref( xml *node, fields *info )
 {
 	int status = BIBL_OK;
 	if ( node->down ) {
-		if ( xml_tagexact( node, "PubmedArticle" ) )
+		if ( xml_tag_matches( node, "PubmedArticle" ) )
 			status = medin_pubmedarticle( node->down, info );
-		else if ( xml_tagexact( node, "MedlineCitation" ) )
+		else if ( xml_tag_matches( node, "MedlineCitation" ) )
 			status = medin_medlinecitation( node->down, info );
 		else
 			status = medin_assembleref( node->down, info );
@@ -610,13 +630,13 @@ medin_assembleref( xml *node, fields *info )
 
 	/* assume everything is a journal article */
 	if ( fields_num( info ) ) {
-		status = fields_add( info, "RESOURCE", "text", 0 );
+		status = fields_add( info, "RESOURCE", "text", LEVEL_MAIN );
 		if ( status!=FIELDS_OK ) return BIBL_ERR_MEMERR;
-		status = fields_add( info, "ISSUANCE", "continuing", 1 );
+		status = fields_add( info, "ISSUANCE", "continuing", LEVEL_HOST );
 		if ( status!=FIELDS_OK ) return BIBL_ERR_MEMERR;
-		status = fields_add( info, "GENRE", "periodical", 1 );
+		status = fields_add( info, "GENRE:MARC", "periodical", LEVEL_HOST );
 		if ( status!=FIELDS_OK ) return BIBL_ERR_MEMERR;
-		status = fields_add( info, "GENRE", "academic journal", 1 );
+		status = fields_add( info, "GENRE:BIBUTILS", "academic journal", LEVEL_HOST );
 		if ( status!=FIELDS_OK ) return BIBL_ERR_MEMERR;
 		status = BIBL_OK;
 	}
@@ -625,13 +645,13 @@ medin_assembleref( xml *node, fields *info )
 }
 
 static int
-medin_processf( fields *medin, char *data, char *filename, long nref, param *p )
+medin_processf( fields *medin, const char *data, const char *filename, long nref, param *p )
 {
 	int status;
 	xml top;
 
 	xml_init( &top );
-	xml_tree( data, &top );
+	xml_parse( data, &top );
 	status = medin_assembleref( &top, medin );
 	xml_free( &top );
 
